/*
 * maebsa.c
 *
 * Copyright 2007 Eric DeVolder
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#undef DEBUG

#include <linux/version.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/poll.h>
#include <linux/bitops.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/hardirq.h>
#include <asm/mipsregs.h>

#include <linux/miscdevice.h>

#include <asm/mach-au1x00/au13xx.h>
#define uint8 unsigned char
#define uint16 unsigned short
#define uint32 unsigned int

#include "maeioctl.h"
#include "bsa.h"

#define AU1300_MAE_BSA_INT 94

#define MODULE_PRETTY_NAME "Alchemy Au13XX MAEBSA"

#define MODULE_NAME  "au13xx_maebsa"
#define VERSION      "1.3"

#define MAEBSA_WATCHDOG_TO_MSECS (200000) /* remove this when timeout is explicitly passed */

#define err(arg...) printk(KERN_ERR MODULE_NAME ": " arg)

static struct maebsa_state {
    int irq;
    uint32 endOfifo0PhysAddr;
    uint32 endOfifo1PhysAddr;
    uint32 endMailbox0;
    uint32 endMailbox1;
    uint32 endMailbox2;
    uint32 endMailbox3;
    uint32 endBitcounter;
    uint32 endMipsCounter;
    wait_queue_head_t wait;
    int status;
    struct semaphore lock;

} maebsa_state;

//////////////////////////////////////////////////////////////////////
static void
maebsa_reset (int reset)
{
    AU13XX_VSSCTRL *vss = (AU13XX_VSSCTRL *)(0xA0000000|VSS_PHYS_ADDR);
    volatile uint32 junk;

    if (reset)
    {
        vss->bsa.clkrst = 3;
    }
    else
    {
        vss->bsa.clkrst = 2;
    }

    // Read-back to force write to go out, provides more time when cycling reset
    junk = vss->bsa.clkrst;
}

//////////////////////////////////////////////////////////////////////
static void
maebsa_power (int up)
{
    AU13XX_VSSCTRL *vss = (AU13XX_VSSCTRL *)(0xA0000000|VSS_PHYS_ADDR);

    if (up)
    {
        maebsa_reset(1);

        vss->bsa.gate = 0x01fffffe;
        vss->bsa.ftr = 1;
        vss->bsa.ftr = 3;
        vss->bsa.ftr = 7;
        vss->bsa.ftr = 0xf;

        vss->bsa.gate = 0x01fffffF;

        maebsa_reset(0);

        vss->bsa.ftr = 0x1f;
    }
    else
    {
        vss->bsa.ftr = 0xF;
        vss->bsa.gate = 0;
        vss->bsa.clkrst = 2;
        vss->bsa.clkrst = 1;
        vss->bsa.ftr = 0;
    }
}

//////////////////////////////////////////////////////////////////////
static void
maebsa_transaction (mae_bsa_request_t *tr)
{
    bsa_t *bsa = (bsa_t *)(0xA0000000|BSA_PHYS_ADDR);
    int timeout, OK = 1;

    tr->bsaStatus = SCB_STATUS_FE; // Default to error state

    // Sanity checks
    if (tr->iramPhysAddr == 0x00000000) OK = 0;
    if (tr->dramPhysAddr == 0x00000000) OK = 0;
    if (tr->pmmPhysAddr == 0x00000000) OK = 0;
    if (tr->ofifo0StartPhysAddr == 0x00000000) OK = 0;
    if (tr->ofifo1StartPhysAddr == 0x00000000) OK = 0;
    if (tr->ofifo0EndPhysAddr == 0x00000000) OK = 0;
    if (tr->ofifo1EndPhysAddr == 0x00000000) OK = 0;
    if (tr->neighPhysAddr == 0x00000000) OK = 0;
    if (tr->descPhysAddr == 0x00000000) OK = 0;
    if (tr->numEntries == 0) OK = 0;
    if (!OK)
    {
        err("problem with BSA request\n");
        return;
    }

    // Acquire lock
    if(down_interruptible(&maebsa_state.lock))
    {
        err("could not acquire lock\n");
        return;
    }

    maebsa_state.status = SCB_STATUS_FE;

    maebsa_reset(1);
    maebsa_reset(0);

    tr->begMipsCounter = read_c0_count();

    //if (tr->flags & MAEBSA_FLAGS_DCWB) dma_cache_wback((unsigned long)tr, 16*1024); // writeback ENTIRE dcache

    // Configure OFIFOs
    bsa->regs.ofifo0.addr.WD32 = tr->ofifo0StartPhysAddr;
    bsa->regs.ofifo0.endaddr.WD32 = tr->ofifo0EndPhysAddr;
    bsa->regs.ofifo1.addr.WD32 = tr->ofifo1StartPhysAddr;
    bsa->regs.ofifo1.endaddr.WD32 = tr->ofifo1EndPhysAddr;
    bsa->regs.ofifo0.control.WD32 = OFIFO_CONTROL_S | tr->ofifo0Control;
    bsa->regs.ofifo1.control.WD32 = OFIFO_CONTROL_S | tr->ofifo1Control;

    // Configure Parsed Macroblock Map
    bsa->regs.scb.mailbox0.WD32 = tr->pmmPhysAddr;

#define GPDMA_M2R (GPDMA_CONTROL_DIRECTION_M2R | GPDMA_CONTROL_START)

    // Load IRAM (using GPDMA)
    bsa->regs.gpdma.ramaddr.WD32 = 0x00008000;
    bsa->regs.gpdma.memaddr.WD32 = tr->iramPhysAddr;
    bsa->regs.gpdma.numwords.WD32 = tr->iramSize;
    bsa->regs.gpdma.control.WD32 = GPDMA_M2R | GPDMA_CONTROL_SD_100;
    // Wait for it to finish
    timeout = 0x00100000;
    do
    {
        if (bsa->regs.gpdma.status.WD32 & GPDMA_STATUS_TC)
            break;
    } while (--timeout > 0);
    if (timeout == 0) { OK = 0; err("IRAM load failed\n"); }

    // Load DRAM (using GPDMA)
    bsa->regs.gpdma.ramaddr.WD32 = 0x00000000;
    bsa->regs.gpdma.memaddr.WD32 = tr->dramPhysAddr;
    bsa->regs.gpdma.numwords.WD32 = tr->dramSize;

    bsa->regs.gpdma.control.WD32 = GPDMA_M2R | tr->gpdmaControl;
    // Wait for it to finish
    timeout = 0x00100000;
    do
    {
        if (bsa->regs.gpdma.status.WD32 & GPDMA_STATUS_TC)
            break;
    } while (--timeout > 0);
    if (timeout == 0) { OK = 0; err("DRAM load failed\n"); }

    // Configure IFIFO
    bsa->regs.ififo.control.WD32 = tr->ififoControl;
    bsa->regs.ififo.pattern.WD32 = tr->ififoPattern;
    bsa->regs.ififo.numentries.WD32 = 0;
    bsa->regs.ififo.listaddr.WD32 = tr->descPhysAddr;
    bsa->regs.ififo.numentries.WD32 = tr->numEntries;

    // Configure Neighbor
    bsa->regs.neigh.spaddr.WD32 = tr->neighPhysAddr;

    // Configure SCB
    bsa->regs.scb.status.WD32 = 0;
    bsa->regs.scb.mailbox1.WD32 = tr->mailbox1;
    bsa->regs.scb.mailbox2.WD32 = tr->mailbox2;
    bsa->regs.scb.mailbox3.WD32 = tr->mailbox3;
    bsa->regs.scb.watchdog.WD32 = tr->watchdog;
    bsa->regs.scb.mask.WD32 = tr->irqMask;
    bsa->regs.scb.addr.WD32 = tr->initPC;

    if (OK)
    {
        // TODO: it would be better to pass the timeout explicitly within the bsa request
        int timeout_msecs, timeout_jiffies;
        timeout_msecs = tr->watchdog / (MAEBSA_WATCHDOG_TO_MSECS);
        if(timeout_msecs == 0)
        {
            timeout_msecs = 5000;
        }

        timeout_jiffies = msecs_to_jiffies(timeout_msecs);

        bsa->regs.scb.go.WD32 = SCB_GO_GO;
        // Now block waiting for BSA to finish - solves race of BSA IRQ occuring before
        // kernel has a chance to put process to sleep (think a few byte slice finishing really fast)
        wait_event_interruptible_timeout(maebsa_state.wait,
                                         (bsa->regs.scb.go.WD32 != SCB_GO_GO),
                                         timeout_jiffies);

        if (maebsa_state.status != SCB_STATUS_DONE)
        {
            // Hang/error, reset BSA
            maebsa_reset(1);
        }
    }

    // Provide feedback from BSA run
    tr->bsaStatus = maebsa_state.status;
    tr->endOfifo0PhysAddr = maebsa_state.endOfifo0PhysAddr;
    tr->endOfifo1PhysAddr = maebsa_state.endOfifo1PhysAddr;
    tr->endMailbox0 =  maebsa_state.endMailbox0;
    tr->endMailbox1 =  maebsa_state.endMailbox1;
    tr->endMailbox2 =  maebsa_state.endMailbox2;
    tr->endMailbox3 =  maebsa_state.endMailbox3;
    tr->endBitcounter = maebsa_state.endBitcounter;

    tr->endMipsCounter = maebsa_state.endMipsCounter;

    // Release lock
    up(&maebsa_state.lock);
}

//////////////////////////////////////////////////////////////////////
irqreturn_t maebsa_handler(int irq, void *dev_id)
{
    bsa_t *bsa = (bsa_t *)(0xA0000000|BSA_PHYS_ADDR);

    //disable_irq(maebsa_state.irq);

    maebsa_state.status = bsa->regs.scb.status.WD32;
    maebsa_state.endOfifo0PhysAddr = bsa->regs.ofifo0.addr.WD32;
    maebsa_state.endOfifo1PhysAddr = bsa->regs.ofifo1.addr.WD32;
    maebsa_state.endMailbox0 = bsa->regs.scb.mailbox0.WD32;
    maebsa_state.endMailbox1 = bsa->regs.scb.mailbox1.WD32;
    maebsa_state.endMailbox2 = bsa->regs.scb.mailbox2.WD32;
    maebsa_state.endMailbox3 = bsa->regs.scb.mailbox3.WD32;
    maebsa_state.endBitcounter = bsa->regs.ififo.bitcounter.WD32;
    maebsa_state.endMipsCounter = read_c0_count();
    bsa->regs.scb.status.WD32 = 0; // ack interrupts

    wake_up_interruptible(&maebsa_state.wait);

    //enable_irq(maebsa_state.irq);

    return IRQ_HANDLED;
}

//////////////////////////////////////////////////////////////////////
static int
maebsa_ioctl(struct inode *inode, struct file *file,
                unsigned int cmd, unsigned long arg)
{
    mae_bsa_request_t *tr = (mae_bsa_request_t *)arg;

    if (tr->magic != MAEBSA_MAGIC)
    {
        err("bad magic %08X\n", tr->magic);
        return -1;
    }

    switch (cmd)
    {
        case MAEBSA_IOCTL_SUBMIT_TRANSACTION:
            maebsa_transaction(tr);
            return 0;
            break;

        default:
            err("error with maebsa request %d\n", cmd);
            return -2;
            break;
    }
}

/*
 * Keep track of the number of times that this device is opened so that we
 * don't turn the power on or off at the wrong time.  Accesses should probably
 * be mutexed but they don't happen too often so we'll be ok.
 */
static int open_count = 0;

//////////////////////////////////////////////////////////////////////
static int
maebsa_open(struct inode *inode, struct file *file)
{
	open_count++;

	if (open_count == 1)
		maebsa_power(1);

    return 0;
}

//////////////////////////////////////////////////////////////////////
static int
maebsa_release(struct inode *inode, struct file *file)
{
	open_count--;

	if (open_count == 0)
		maebsa_power(0);

    return 0;
}

//////////////////////////////////////////////////////////////////////
static struct file_operations maebsa_fops =
{
    owner:      THIS_MODULE,
    open:       maebsa_open,
    ioctl:      maebsa_ioctl,
    release:    maebsa_release,
};

static struct miscdevice maebsa_device = {
    .name = "maebsa",
    .minor = MISC_DYNAMIC_MINOR,
    .fops = &maebsa_fops,
};

//////////////////////////////////////////////////////////////////////
static int __init
maebsa_init(void)
{
    int rcode = 0;
    maebsa_state.irq = AU1300_MAE_BSA_INT;
    if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,11))
        maebsa_state.irq += 8;

    init_waitqueue_head(&maebsa_state.wait);
    sema_init(&maebsa_state.lock, 1);

    /* Register Device*/
    rcode = misc_register(&maebsa_device);
    if ( rcode < 0 )
        err("Could not register MAE BSA device\n");
    else
    {
        if ( (rcode=request_irq(maebsa_state.irq, maebsa_handler,
                                IRQF_SHARED, MODULE_NAME, (void *)&maebsa_state )) )
        {
            err("Could not get IRQ %d", maebsa_state.irq);
        }
        else
        {
            printk(MODULE_PRETTY_NAME " " ENDIAN_STR " driver registered Successfully v%s (magic %x)\n", VERSION, MAEBSA_MAGIC);
        }
    }

    return rcode;
}

//////////////////////////////////////////////////////////////////////
static void __exit
maebsa_exit(void)
{
    /*Unregister Device*/
    misc_deregister(&maebsa_device);
    free_irq(maebsa_state.irq, (void *)&maebsa_state);

}

//////////////////////////////////////////////////////////////////////

MODULE_AUTHOR("Eric DeVolder");
MODULE_DESCRIPTION("Au13XX MAE BSA Driver");
MODULE_LICENSE("GPL");
module_init(maebsa_init);
module_exit(maebsa_exit);

