/*
 * maeite.c
 *
 * Copyright 2007 Eric DeVolder
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#undef DEBUG

#include <linux/version.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/poll.h>
#include <linux/bitops.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/dma-mapping.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/hardirq.h>
#include <asm/mipsregs.h>

#include <linux/miscdevice.h>

#define uint8 unsigned char
#define uint16 unsigned short
#define uint32 unsigned int

#include "ite.h"
#include "maeioctl.h"

#if defined(AU1200) || defined(CONFIG_SOC_AU1200)
#define AU1200_MAE_BE_INT 4 // FIX!!! from asm/mach-au1x00/au1000.h
#define MODULE_NAME  "au12xx_maebe"
#define MODULE_PRETTY_NAME "Alchemy Au1200 MAEBE"
#define MISC_NAME "maebe"
#define MAEITE_MAX_FRAME_MS 200
#endif

#if defined(AU13XX)
#define AU1200_MAE_BE_INT 92
#define MODULE_NAME  "au13xx_maeite"
#define MODULE_PRETTY_NAME "Alchemy Au13XX MAEITE"
#define MISC_NAME "maeite"
#define MAEITE_MAX_FRAME_MS 100
#endif

#define VERSION     "1.3"

#define err(arg...) printk(KERN_ERR MODULE_NAME ": " arg)

//////////////////////////////////////////////////////////////////////
static struct maebe_state {
    int irq;
    uint32 endMipsCounter;
    wait_queue_head_t wait;
    volatile int done;
    struct semaphore lock;

} maebe_state;

//////////////////////////////////////////////////////////////////////
static void
be_init (void)
{
    // Initialize MAEBE
    AU1200_MAEBE * const be = (AU1200_MAEBE *)(0xA0000000|MAEBE_PHYS_ADDR);

    be->ctlenable    = MAEBE_CTLENABLE_EN; // enable clocks to BE
    be->ctlfpc       = MAEBE_CTLFPC_FRST; // reset the BE
}

//////////////////////////////////////////////////////////////////////
static void
be_deinit (void)
{
    AU1200_MAEBE * const be = (AU1200_MAEBE *)(0xA0000000|MAEBE_PHYS_ADDR);

    be->ctlenable = 0; // disable clocks
}

//////////////////////////////////////////////////////////////////////
static void
be_submit (mae_be_request_t *tr)
{
    AU1200_MAEBE * const be = (AU1200_MAEBE *)(0xA0000000|MAEBE_PHYS_ADDR);
    int i;

#ifdef CONFIG_DMA_COHERENT
    if (tr->flags & MAEBE_FLAGS_DCWB) dma_cache_wback((unsigned long)tr, 16*1024); // writeback ENTIRE dcache
#endif

    // Acquire lock
    if (down_interruptible(&maebe_state.lock))
    {
        err("Could not acquire lock\n");
        return;
    }

    tr->begMipsCounter = read_c0_count();

    be->ctlenable = 0; // reset block
    (void)be->ctlenable; // force writes to go out
    be->ctlenable = MAEBE_CTLENABLE_EN; // enable block
    (void)be->ctlenable; // force writes to go out
    be->ctlfpc = MAEBE_CTLFPC_FRST; // reset for next transaction
    (void)be->ctlenable; // force writes to go out

    for (i = 0; i < 32; ++i)
    {
        be->scfhalut[i] = tr->scfhalut[i];
        be->scfvalut[i] = tr->scfvalut[i];
        be->scfhblut[i] = tr->scfhblut[i];
        be->scfvblut[i] = tr->scfvblut[i];
        be->scfhclut[i] = tr->scfhclut[i];
        be->scfvclut[i] = tr->scfvclut[i];
    }

    be->cscxcffa = tr->cscxcffa;
    be->cscxcffb = tr->cscxcffb;
    be->cscxcffc = tr->cscxcffc;
    be->cscycffa = tr->cscycffa;
    be->cscycffb = tr->cscycffb;
    be->cscycffc = tr->cscycffc;
    be->csczcffa = tr->csczcffa;
    be->csczcffb = tr->csczcffb;
    be->csczcffc = tr->csczcffc;
    be->cscxoff  = tr->cscxoff;
    be->cscyoff  = tr->cscyoff;
    be->csczoff  = tr->csczoff;
// FIX!!! missing cscalpha

    be->scfhsr   = tr->scfhsr;
    be->scfvsr   = tr->scfvsr;
    be->scfdisable = tr->scfdisable;

    be->srccfg = tr->srccfg;
    be->srcfhw = tr->srcfhw;
    be->srcaaddr = tr->srcaaddr;
    be->srcbaddr = tr->srcbaddr;
    be->srccaddr = tr->srccaddr;
    be->srcastr = tr->srcastr;
    be->srcbstr = tr->srcbstr;
    be->srccstr = tr->srccstr;

    be->dstaddr = tr->dstaddr;
    be->dststr = tr->dststr;
    be->dstheight = tr->dstheight;
    be->dstcfg = tr->dstcfg;

    be->ctlintstat = ~0;
    be->ctlintenable = MAEBE_CTLINTSTAT_FC;

    maebe_state.done = 0;
    be->ctlfpc = MAEBE_CTLFPC_STR;
    wait_event_interruptible_timeout(maebe_state.wait,
                                     (maebe_state.done == 1),
                                     msecs_to_jiffies(MAEITE_MAX_FRAME_MS));
    if(! maebe_state.done)
    {
        be->ctlfpc = MAEBE_CTLFPC_FRST; // reset for next transaction

        err("hang\n");
    }

    tr->endMipsCounter = maebe_state.endMipsCounter;

    // Release lock
    up(&maebe_state.lock);
}

//////////////////////////////////////////////////////////////////////
irqreturn_t maebe_handler(int irq, void *dev_id
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,11))
    , struct pt_regs *regs
#endif
)
{
    AU1200_MAEBE * const be = (AU1200_MAEBE *)(0xA0000000|MAEBE_PHYS_ADDR);

    //disable_irq(maebe_state.irq);

    be->ctlintstat = ~0; // ack interrupts

    maebe_state.endMipsCounter = read_c0_count();

    maebe_state.done = 1;
    wake_up_interruptible(&maebe_state.wait);

    //enable_irq(maebe_state.irq);

    return IRQ_HANDLED;
}

//////////////////////////////////////////////////////////////////////
static int
maebe_ioctl(struct inode *inode, struct file *file,
                unsigned int cmd, unsigned long arg)
{
    mae_be_request_t *tr = (mae_be_request_t *)arg; // FIX ideally need to copy_from_user(), but not necessary on mips

    if (tr->magic != MAEBE_MAGIC)
    {
        err("bad magic %08X\n", tr->magic);
        return -1;
    }

    switch (cmd)
    {
        case MAEBE_IOCTL_SUBMIT_TRANSACTION:
            be_submit(tr);
            break;
        default:
            err("error with maebe request %d\n", cmd);
            break;
    }
    return 0;
}

//////////////////////////////////////////////////////////////////////
static int
maebe_open(struct inode *inode, struct file *file)
{
    return 0;
}

//////////////////////////////////////////////////////////////////////
static int
maebe_release(struct inode *inode, struct file *file)
{
    return 0;
}

//////////////////////////////////////////////////////////////////////
static struct file_operations maebe_fops =
{
    owner:      THIS_MODULE,
    open:       maebe_open,
    ioctl:      maebe_ioctl,
    release:    maebe_release,
};

static struct miscdevice maebe_device = {
    .name = MISC_NAME,
    .minor = MISC_DYNAMIC_MINOR,
    .fops = &maebe_fops,
};

//////////////////////////////////////////////////////////////////////
static int __init
maebe_init(void)
{
    int rcode = 0;

    maebe_state.irq = AU1200_MAE_BE_INT;
    if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,11))
        maebe_state.irq += 8;

    init_waitqueue_head(&maebe_state.wait);
    sema_init(&maebe_state.lock, 1);

    /* Register Device*/
    rcode = misc_register(&maebe_device);
    if ( rcode < 0 )
        err("Could not register device\n");
    else
    {
        if ( (rcode=request_irq(maebe_state.irq, maebe_handler,
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,11))
                            SA_SHIRQ | SA_INTERRUPT,
#else
                            IRQF_SHARED,
#endif
                            MODULE_NAME, (void *)&maebe_state )) )
        {
            err("Could not get IRQ %d", maebe_state.irq);
        }
        else
        {
            be_init();
            printk(MODULE_PRETTY_NAME " " ENDIAN_STR " driver registered Successfully v%s (magic %x)\n", VERSION, MAEBE_MAGIC);
        }
    }

    return rcode;
}

//////////////////////////////////////////////////////////////////////
static void __exit
maebe_exit(void)
{
    be_deinit();
    misc_deregister(&maebe_device);
    free_irq(maebe_state.irq, (void *)&maebe_state);
}

MODULE_AUTHOR("Eric DeVolder");
MODULE_DESCRIPTION(MODULE_PRETTY_NAME " Driver");
MODULE_LICENSE("GPL");
module_init(maebe_init);
module_exit(maebe_exit);

