/*
 * maempe.c
 *
 * Copyright 2007 Eric DeVolder
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#undef DEBUG

#include <linux/version.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/poll.h>
#include <linux/bitops.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/hardirq.h>
#include <asm/mipsregs.h>

#include <linux/miscdevice.h>

#include <asm/mach-au1x00/au13xx.h>
#define uint8 unsigned char
#define uint16 unsigned short
#define uint32 unsigned int

#include "maeioctl.h"
#include "mpe.h"

#define AU1300_MAE_MPE_INT 93

#define MODULE_PRETTY_NAME "Alchemy Au13XX MAEMPE"

#define MODULE_NAME  "au13xx_maempe"
#define VERSION      "1.4"

// FIXME: pass a timeout argument instead of this hardcoded value
#define MAEMPE_MAX_FRAME_MS 500

#define err(arg...) printk(KERN_ERR MODULE_NAME ": " arg)

#define INT_MODULE_PARM(n, v) static int n = v; module_param(n, int, 0)
INT_MODULE_PARM(pri0, 0x45617238); // hw default
INT_MODULE_PARM(pri1, 0x09000000); // hw default

static struct maempe_state {
    int irq;
    uint32 endMipsCounter;
    wait_queue_head_t wait;
    volatile int done;
    struct semaphore lock;
    uint32 prev_cfg8;
} maempe_state;

enum
{
    MAEMPE_OK = 0,
    MAEMPE_ERROR_MAGIC,
    MAEMPE_ERROR_HANG,
    MAEMPE_ERROR_CMD,
} maempe_result;

//////////////////////////////////////////////////////////////////////
static void
maempe_reset (int reset)
{
    AU13XX_VSSCTRL *vss = (AU13XX_VSSCTRL *)(0xA0000000|VSS_PHYS_ADDR);
    mpe_dma_t *mpe_dma = (mpe_dma_t *) (0xA0000000 | MPE_DMA_PHYS_ADDR);

    if (reset)
    {
        vss->mpe.clkrst = 3;
    }
    else
    {
        vss->mpe.clkrst = 2;

        // Junk read to delay for a few cycles, while reset deassertion is propagated.
        {
            int tmp;
            int i;

            for(i = 0; i < 5; i++)
            {
                tmp += mpe_dma->pri0;
            }
        }

        mpe_dma->pri0 = pri0;
        mpe_dma->pri1 = pri1;
    }
}

//////////////////////////////////////////////////////////////////////
static void
maempe_power (int up)
{
    AU13XX_VSSCTRL *vss = (AU13XX_VSSCTRL *)(0xA0000000|VSS_PHYS_ADDR);

    if (up)
    {
        maempe_reset(1);

        vss->mpe.gate = 0x01fffffe;
        vss->mpe.ftr = 1;
        vss->mpe.ftr = 3;
        vss->mpe.ftr = 7;
        vss->mpe.ftr = 0xf;

        vss->mpe.gate = 0x01fffffF;

        maempe_reset(0);

        vss->mpe.ftr = 0x1f;
    }
    else
    {
        vss->mpe.ftr = 0xF;
        vss->mpe.gate = 0;
        vss->mpe.clkrst = 2;
        vss->mpe.clkrst = 1;
        vss->mpe.ftr = 0;
    }
}

//////////////////////////////////////////////////////////////////////
static int
maempe_transaction (mae_mpe_request_t *tr)
{
    mpe_t *mpe = (mpe_t *)(0xA0000000|MPE_PHYS_ADDR);
    int result = MAEMPE_OK;

#ifdef CONFIG_DMA_COHERENT
    if (tr->flags & MAEMPE_FLAGS_DCWB) dma_cache_wback((unsigned long)tr, 16*1024); // writeback ENTIRE dcache
#endif

    // Acquire lock
    if(down_interruptible(&maempe_state.lock))
    {
        err("could not acquire lock\n");
        return MAEMPE_ERROR_HANG;
    }

    tr->begMipsCounter = read_c0_count();

    // For now, only reset the MPE if the codec changes, as jpeg multi-segment decode has internal state that needs
    // to be saved in between passes.
    if (tr->cfg8 != maempe_state.prev_cfg8)
    {
        // Toggle reset before each frame.
        // Especially needed when doing dual-decode with different codecs.
        maempe_reset(1);
        maempe_state.prev_cfg8 = tr->cfg8;
    }

    maempe_reset(0);

#define WRREG(REG) { /*printk( #REG ": %08X\n", tr->REG);*/ mpe->REG = tr->REG; /*printk(#REG ": %08X\n", mpe->REG);*/ }

    maempe_state.done = 0;

    // Write the registers
    WRREG(cfg1);
    WRREG(cfg2);
    WRREG(cfg3);
    WRREG(cfg4);
    WRREG(cfg5);
    WRREG(cfg6);
    WRREG(cfg7);
    WRREG(cfg8);
    WRREG(cfg9);
    WRREG(cfg10);
    WRREG(cfg11);
    WRREG(cfg12);
    WRREG(cfg13);
    WRREG(cfg14);
    WRREG(cfg15);
    WRREG(cfg18);

    mpe->cfg19 = 1; // interrupt mask enable

    mpe->cfg0 = MPE_CFG0_START; // start MPE
    wait_event_interruptible_timeout(maempe_state.wait,
                                     (maempe_state.done == 1),
                                     msecs_to_jiffies(MAEMPE_MAX_FRAME_MS));

    if(! maempe_state.done)
    {
        // Hang, reset MPE
        maempe_reset(1);

        err("hang\n");

        result = MAEMPE_ERROR_HANG;
    }

    tr->endMipsCounter = maempe_state.endMipsCounter;

    // Release lock
    up(&maempe_state.lock);

    return result;
}

//////////////////////////////////////////////////////////////////////
irqreturn_t maempe_handler(int irq, void *dev_id)
{
    //disable_irq(maempe_state.irq);

    maempe_state.endMipsCounter = read_c0_count();

    // Edge interrupt, no need to ack
    maempe_state.done = 1;
    wake_up_interruptible(&maempe_state.wait);

    //enable_irq(maempe_state.irq);

    return IRQ_HANDLED;
}

//////////////////////////////////////////////////////////////////////
static int
maempe_ioctl(struct inode *inode, struct file *file,
             unsigned int cmd, unsigned long arg)
{
    mae_mpe_request_t *tr = (mae_mpe_request_t *)arg;
    int result;

    if (tr->magic != MAEMPE_MAGIC)
    {
        err("bad magic %08X, expected %08X\n", tr->magic, MAEMPE_MAGIC);
        result = MAEMPE_ERROR_MAGIC;
    }
    else
    {
        switch (cmd)
        {
            case MAEMPE_IOCTL_SUBMIT_TRANSACTION:
                result = maempe_transaction(tr);
                break;

            default:
                err("Invalid maempe request %d\n", cmd);
                result = MAEMPE_ERROR_CMD;
                break;
        }
    }

    return result;
}

/*
 * Keep track of the number of times that this device is opened so that we
 * don't turn the power on or off at the wrong time.  Accesses should probably
 * be mutexed but they don't happen too often so we'll be ok.
 */
static int open_count = 0;

//////////////////////////////////////////////////////////////////////
static int
maempe_open(struct inode *inode, struct file *file)
{
	open_count++;

	if (open_count == 1)
		maempe_power(1);

    return 0;
}

//////////////////////////////////////////////////////////////////////
static int
maempe_release(struct inode *inode, struct file *file)
{
	open_count--;

	if (open_count == 0)
		maempe_power(0);

    return 0;
}

//////////////////////////////////////////////////////////////////////
static struct file_operations maempe_fops =
{
    owner:      THIS_MODULE,
    open:       maempe_open,
    ioctl:      maempe_ioctl,
    release:    maempe_release,
};

//////////////////////////////////////////////////////////////////////

static struct miscdevice maempe_device = {
	.name = "maempe",
	.minor = MISC_DYNAMIC_MINOR,
	.fops = &maempe_fops,
};

static int __init
maempe_init(void)
{
    int rcode = 0;

    memset(&maempe_state, 0, sizeof(maempe_state));

    maempe_state.irq = AU1300_MAE_MPE_INT;
    if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,11))
        maempe_state.irq += 8;

    init_waitqueue_head(&maempe_state.wait);
    sema_init(&maempe_state.lock, 1);

    /* Register Device*/
	rcode = misc_register(&maempe_device);
	if (rcode) {
		err("Failed to register\n");
		return rcode;
	}

    if ( (rcode=request_irq(maempe_state.irq, maempe_handler,
                            IRQF_SHARED, MODULE_NAME, (void *)&maempe_state )) )
    {
        err("Could not get IRQ %d", maempe_state.irq);
    }
    else
    {
        printk(MODULE_PRETTY_NAME " " ENDIAN_STR " driver registered Successfully v%s (magic %x)\n", VERSION, MAEMPE_MAGIC);
        printk(MODULE_PRETTY_NAME " mae_dma: pri0 %08X, pri1 %08X\n", pri0, pri1);
    }

    return rcode;
}

//////////////////////////////////////////////////////////////////////
static void __exit
maempe_exit(void)
{
    /*Unregister Device*/
    misc_deregister(&maempe_device);
    free_irq(maempe_state.irq, (void *)&maempe_state);
}

//////////////////////////////////////////////////////////////////////

MODULE_AUTHOR("Eric DeVolder");
MODULE_DESCRIPTION(MODULE_PRETTY_NAME " Driver");
MODULE_LICENSE("GPL");
module_init(maempe_init);
module_exit(maempe_exit);

